# Copyright (c) 2025, InfinityQ Technology, Inc.

from typing import List, Optional
from typing_extensions import override, Self

from titanq._event.sink import ContentBuilder, DefinedProgress, Sink, UndefinedProgress
from titanq._util.decorators import chainable



class NoOpContentBuilder(ContentBuilder[str]):

    @override
    def build(self) -> str: ...

    @override
    @chainable
    def set_color(self, color: str) -> Self: ...

    @override
    @chainable
    def add_emoji(self, emoji: str) -> Self: ...

    @override
    @chainable
    def add_text(self, text: str) -> Self:...

    @override
    @chainable
    def add_whitespace(self) -> Self: ...

    @override
    @chainable
    def add_tailing_spinner(self) -> Self: ...


class NoOpDefinedProgress(DefinedProgress[str]):

    @override
    def add_task(self, renderable: str, total: int) -> int: ...

    @override
    def content(self) -> str: ...

    @override
    def update(self, handle: int, new_value: int, renderable: Optional[str] = None) -> None: ...

    @override
    def finalize(self) -> None: ...


class NoOpUndefinedProgress(UndefinedProgress[str]):

    def update(self, renderable: str, additional_renderables: Optional[List[str]] = None) -> str: ...


class NoOpSink(Sink):
    """A sink that output nothing"""

    @override
    def start(self) -> None:
        pass

    @override
    def stop(self) -> None:
        pass

    @override
    def display(self, renderable: str, with_timestamp: Optional[bool] = True) -> None:
        pass

    @override
    def content_builder(self) -> ContentBuilder[str]:
        return NoOpContentBuilder()

    @override
    def defined_progress(self) -> DefinedProgress[str]:
        return NoOpDefinedProgress()

    @override
    def undefined_progress(self) -> UndefinedProgress[str]:
        return NoOpUndefinedProgress()


# Copyright (c) 2025, InfinityQ Technology, Inc.
from enum import Enum

import numpy as np

from titanq._model.array.arraylike.numpy_array import NumpyArray



class ConstraintType(str, Enum):
    SET_PARTITIONING = "set_partitioning"
    CARDINALITY = "cardinality"
    EQUALITY = "equality"
    INEQUALITY = "inequality"


def get_constraint_type(mask: NumpyArray, bounds: NumpyArray) -> ConstraintType:
    """Returns the type of the constraint using the mask and bounds.

    :param mask: 1D Constraint mask
    :param bounds: 1D Constraint bounds

    :raises ValueError: If the shape of the mask or bounds doesn't match

    :return: The constraint type
    :rtype: ConstraintType
    """
    if bounds.ndim() != 1 or bounds.shape()[0] != 2:
        raise ValueError(f"constraint bounds has an invalid shape: {bounds.shape()}, expected: (2,)")


    if mask.ndim() != 1:
        raise ValueError(f"constraint mask has an invalid dimension: {mask.ndim()}, expected: 1")

    # use directly the numpy data
    bounds = bounds.inner()

    # detect inequality constraints
    if any(np.isnan(bounds)) or bounds[0] != bounds[1]:
        return ConstraintType.INEQUALITY

    # both bounds are equal from now on

    # any non-binary is an equality
    if not mask.is_binary():
        return ConstraintType.EQUALITY

    # any non-integer bounds or lower than 1 is an equality
    if bounds[0] % 1 != 0 or bounds[0] < 1:
        return ConstraintType.EQUALITY

    # equality value is equal to one
    if bounds[0] == 1:
        return ConstraintType.SET_PARTITIONING

    # equality value is lower or equal to the sum of the mask elements
    if bounds[0] <= mask.sum():
        return ConstraintType.CARDINALITY
    else:
        return ConstraintType.EQUALITY



# Copyright (c) 2025, InfinityQ Technology, Inc.

from typing import Dict, List, Optional
from typing_extensions import override, Self

import logging

from titanq._event.sink import ContentBuilder, DefinedProgress, Sink, UndefinedProgress
from titanq._util.decorators import chainable


# Configure the logger
logging.basicConfig(
    format='[%(asctime)s] %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=logging.INFO
)


class TextContentBuilder(ContentBuilder[str]):

    def __init__(self):
        self._content = ""

    @override
    def build(self) -> str:
        return self._content.strip()

    @override
    @chainable
    def set_color(self, color: str) -> Self:
        # no coloring for text based output
        pass

    @override
    @chainable
    def add_emoji(self, emoji: str) -> Self:
        # no emoji for text based output
        pass

    @override
    @chainable
    def add_text(self, text: str) -> Self:
        self._content += text

    @override
    @chainable
    def add_whitespace(self) -> Self:
        self._content += " "

    @override
    @chainable
    def add_tailing_spinner(self) -> Self:
        # no spinner animation for text based output
        pass


class TextDefinedProgress(DefinedProgress[str]):

    def __init__(self):
        self._progress_tasks: Dict[int, int] = {}
        self._content = None

    @override
    def add_task(self, renderable: str, total: int) -> int:
        self._content = renderable
        return 0 # unused, just return 0

    @override
    def content(self) -> str:
        return self._content

    @override
    def update(self, handle: int, new_value: int, renderable: Optional[str] = None) -> None:
        pass

    @override
    def finalize(self) -> None:
        pass


class TextUndefinedProgress(UndefinedProgress[str]):

    def __init__(self):
        # keep track if a log was already logged
        self._already_logged = []

    @override
    def update(self, renderable: str, additional_renderables: Optional[List[str]] = None) -> str:
        if renderable not in self._already_logged:
            self._already_logged.append(renderable)
            return renderable

        if additional_renderables:
            for renderable in additional_renderables:
                if renderable not in self._already_logged:
                    self._already_logged.append(renderable)
                    return renderable

        return ""


class TextSink(Sink[str]):
    """The TextSink is just a simple python logger."""

    def __init__(
        self,
        version: str,
        show_header: bool,
        start_separator: Optional[str] = None
    ):
        """
        Initialize a simple text sink logger.

        :param version: Version shown in the sink
        :param show_header: If True, will show a header in the startup
        :param start_separator: Separator text to include if set that will be shown in the startup
        """
        self._version = version
        self._show_header = show_header
        self._start_separator = start_separator

    @override
    def start(self) -> None:
        if self._start_separator is not None:
            logging.info(f"Using TitanQ SDK {self._version}")
        logging.info(f"------------------------ {self._start_separator} ------------------------")

    @override
    def stop(self) -> None:
        pass

    @override
    def display(self, renderable: str, with_timestamp: Optional[bool] = True) -> None:
        # if nothing, do not display
        if renderable != "":
            logging.info(renderable)

    @override
    def content_builder(self) -> ContentBuilder[str]:
        return TextContentBuilder()

    @override
    def defined_progress(self) -> DefinedProgress[str]:
        return TextDefinedProgress()

    @override
    def undefined_progress(self) -> UndefinedProgress[str]:
        return TextUndefinedProgress()


# Copyright (c) 2025, InfinityQ Technology, Inc.

from typing import Type, Union

import numpy as np
from scipy.sparse import coo_array, csr_array

from titanq._model.array import _TITANQ_DTYPE, Array, conversion_util
from titanq._model.array.arraybuilder import ArrayBuilder
from titanq._model.array.arraylike import ArrayLike
from titanq._model.array.arraylike.numpy_array import NumpyArray
from titanq._model.array.arraylike.scipy_coo_array import ScipyCooArray
from titanq._model.array.arraylike.scipy_csr_array import ScipyCsrArray
from titanq._model.array.transform import ArrayTransformChecker
from titanq._model.array.transform.sparse_threshold import SparseThresholdStrategy


_SPARSE_THRESHOLD = 0.2 # 20%

ArrayLikeFactoryInputs = Union[Array, ArrayBuilder]


class ArrayLikeFactory:
    """
    `ArrayLikeFactory` handles the creation of ArrayLike objects in the array module.

    Methods
    -------
    `create_numpy()` will create a numpy array like object.

    `create_scipy_coo_array()` will create a scipy coo array like object.

    `create_minimal()` will create a resource efficient array like object. It focuses
    on minimizing memory usage.
    """
    def __init__(self):
        self._sparse_transform_checker = ArrayTransformChecker(SparseThresholdStrategy(_SPARSE_THRESHOLD))

    def create_numpy(self, array: ArrayLikeFactoryInputs, data_type: np.dtype = _TITANQ_DTYPE) -> NumpyArray:
        """Creates a `NumpyArray` array like object."""
        array_builder = self._as_array_builder(array, data_type)

        return self._array_like_from_array_builder(array_builder, NumpyArray)

    def create_minimal(self, array: ArrayLikeFactoryInputs) -> ArrayLike:
        """
        Returns the corresponding `ArrayLike` based on the instance type, optimizing for
        minimal memory usage.
        """
        array_builder = self._as_array_builder(array)

        # apply the sparse threshold strategy
        array_like_type = self._sparse_transform_checker.check(array_builder)

        return self._array_like_from_array_builder(array_builder, array_like_type)

    def _as_array_builder(self, array: ArrayLikeFactoryInputs, data_type: np.dtype = _TITANQ_DTYPE) -> ArrayBuilder:
        """Creates and returns the `ArrayBuilder` based on the instance type."""
        if isinstance(array, np.ndarray):
            return NumpyArray(array.astype(data_type))

        elif isinstance(array, coo_array):
            return ScipyCooArray(array.astype(data_type))

        elif isinstance(array, csr_array):
            return ScipyCsrArray(array.astype(data_type))

        elif isinstance(array, ArrayBuilder): # Already an array builder
            return array

        else:
            raise ValueError(f"Unsupported array type: {type(array)}")

    def _array_like_from_array_builder(self, array: ArrayBuilder, type: Type[ArrayLike]) -> ArrayLike:
        """Converts an `ArrayBuilder` into an `ArrayLike` depending on the provided `type`."""
        if isinstance(array, type):
            return array # there is no need to convert it to itself

        if type is NumpyArray:
            numpy_array = conversion_util.array_builder_to_numpy(array)
            return self._as_array_builder(numpy_array)

        elif type is ScipyCooArray:
            coo_array = conversion_util.array_builder_to_coo_array(array)
            return self._as_array_builder(coo_array)

        elif type is ScipyCsrArray:
            csr_array = conversion_util.array_builder_to_csr_array(array)
            return self._as_array_builder(csr_array)



# Copyright (c) 2025, InfinityQ Technology, Inc.

from io import BytesIO
from typing import Callable
from typing_extensions import override

class BytesReaderWithCallback(BytesIO):

    def __init__(self, initial_bytes = b""):
        super().__init__(initial_bytes)
        self._callback_update = None
        self._registered_total_size = 0

    def set_callback(self, c: Callable[['BytesReaderWithCallback'], None] = None) -> None:
        self._callback_update = c

    @override
    def read(self, size = -1):
        data = super().read(size)
        if self._callback_update:
            self._callback_update(self)
        return data

    @override
    def write(self, buffer):
        data = super().write(buffer)
        if self._callback_update:
            self._callback_update(self)
        return data

    @override
    def close(self) -> None:
        # keep the latest total size before closing
        self._registered_total_size = self.getbuffer().nbytes
        super().close()

    def total_size(self) -> int:
        """Return the total number of bytes in the buffer."""
        return self._registered_total_size if self.closed else self.getbuffer().nbytes


# Copyright (c) 2025, InfinityQ Technology, Inc.

import abc
from io import BytesIO
from typing import Any, Dict, Iterable

import boto3
import boto3.s3
import boto3.s3.transfer
from google.cloud import storage
from google.oauth2 import service_account
from pydantic import BaseModel, ConfigDict, SerializerFunctionWrapHandler, model_serializer
import requests

from titanq._api.model._util import SecretJson, SecretStr
from titanq._api.model.temp_storage import DualUrl


_DOWNLOAD_CHUNK_SIZE = 8192
_TRANSFER_CONFIG = boto3.s3.transfer.TransferConfig(
    multipart_chunksize=5 * 1024 * 1024, # 5 MB
    multipart_threshold=5 * 1024 * 1024, # 5 MB
)


class Location(BaseModel, abc.ABC):
    """ A Location object enables a user to download data from a certain service, uploading is optional """

    @abc.abstractmethod
    def download(self, key: str, bytes_reader: BytesIO) -> None:
        """ download data from a key """

    @abc.abstractmethod
    def download_content_length(self, key: str) -> int:
        """ returns the size of the content of a key to download"""

    @abc.abstractmethod
    def upload(self, key: str, data: BytesIO) -> None:
        """ uploads data to a key, will return if succeeded """


class S3Location(Location):
    """ Simple Storage Service (S3) storage is part of AWS (Amazon Web Services) """
    bucket_name: str
    access_key_id: str
    secret_access_key: SecretStr

    @model_serializer(mode='wrap')
    def serialize_model(self, nxt: SerializerFunctionWrapHandler) -> Dict[str, Any]:
        return {"s3": nxt(self)}

    def _s3_client(self):
        return boto3.client('s3', aws_access_key_id=self.access_key_id, aws_secret_access_key=self.secret_access_key)

    def download(self, key: str, bytes_reader: BytesIO) -> None:
        client = self._s3_client()
        remote_object = client.get_object(Bucket=self.bucket_name, Key=key)
        body = remote_object['Body']

        while True:
            chunk = body.read(_DOWNLOAD_CHUNK_SIZE)
            if not chunk:
                break
            bytes_reader.write(chunk)

        bytes_reader.seek(0)

    def upload(self, key: str, data: BytesIO):
        client = self._s3_client()
        client.upload_fileobj(data, Bucket=self.bucket_name, Key=key, Config=_TRANSFER_CONFIG)

    def download_content_length(self, key: str) -> int:
        client = self._s3_client()
        response = client.head_object(Bucket=self.bucket_name, Key=key)
        return response['ContentLength']


class GcpLocation(Location):
    """ Google Clous Storage (GCS) is part of GCP (Google Cloud Provider) """
    model_config = ConfigDict(validate_assignment=True)

    bucket_name: str
    json_key: SecretJson

    @model_serializer(mode='wrap')
    def serialize_model(self, nxt: SerializerFunctionWrapHandler) -> Dict[str, Any]:
        return {"gcp": nxt(self)}

    def download(self, key: str, bytes_reader: BytesIO) -> None:
        credentials = service_account.Credentials.from_service_account_info(self.json_key.non_obfuscated_value())
        client = storage.Client(credentials=credentials)
        bucket = client.get_bucket(self.bucket_name)

        blob = bucket.blob(key)
        blob.download_to_file(bytes_reader)
        bytes_reader.seek(0)

    def upload(self, key: str, data: BytesIO):
        raise NotImplementedError(f"Uploading is not currently supported by {self.__class__.__name__}")

    def download_content_length(self, key: str) -> int:
        raise NotImplementedError(f"Obtaining content length is not currenty supported by {self.__class__.__name__}")


class UrlLocation(Location):
    """ Download or upload using an URL """

    @model_serializer(mode='wrap')
    def serialize_model(self, _: SerializerFunctionWrapHandler) -> Dict[str, Any]:
        return { "file_name_is_url": True }

    def download(self, key: str, bytes_reader: BytesIO) -> None:
        with requests.get(key, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=_DOWNLOAD_CHUNK_SIZE):
                if chunk:
                    bytes_reader.write(chunk)

            bytes_reader.seek(0)

    def upload(self, key: str, data: BytesIO):
        response = requests.put(key, data=data)
        response.raise_for_status()

    def download_content_length(self, key: str) -> int:
        response = requests.head(key)
        return int(response.headers.get('Content-Length'))


class TempStorageLocation(UrlLocation):
    """ Temp storage location utilize the temp storage service """
    _url_mapping: Dict[str, DualUrl]

    def __init__(self, dual_urls: Iterable[DualUrl], **kwargs):
        super().__init__(**kwargs)
        self._url_mapping = {}
        for dual_url in dual_urls:
            self._url_mapping[dual_url.upload] = dual_url
            self._url_mapping[dual_url.download] = dual_url

    def download(self, key: str, bytes_reader: BytesIO) -> None:
        download_url = self._url_mapping[key].download
        return super().download(download_url, bytes_reader)

    def download_content_length(self, key: str) -> int:
        download_url = self._url_mapping[key].download
        return super().download_content_length(download_url)

    def upload(self, key: str, data: BytesIO):
        upload_url = self._url_mapping[key].upload
        super().upload(upload_url, data)



class BitBucketLocation(Location):
    """ Bitbucket is a git version control, use it to download files from it """
    repo_name: str
    workspace: str
    access_token: SecretStr
    branch: str

    @model_serializer(mode='wrap')
    def serialize_model(self, nxt: SerializerFunctionWrapHandler) -> Dict[str, Any]:
        return {"bitbucket": nxt(self)}

    def download(self, key: str, bytes_reader: BytesIO) -> None:
        url = self._build_url(key)
        response = self._http_get(url)
        bytes_reader.write(response.content)
        bytes_reader.seek(0)

    def upload(self, key: str, data: BytesIO):
        raise NotImplementedError(f"Uploading is not currently supported by {self.__class__.__name__}")

    def download_content_length(self, key: str) -> int:
        raise NotImplementedError(f"Obtaining content length is not currenty supported by {self.__class__.__name__}")

    def _http_get(self, url: str) -> requests.Response:
        return requests.get(url, headers={"Authorization": f"Bearer {self.access_token.non_obfuscated_value()}"})

    def _build_url(self, key: str) -> str:
        return f"https://api.bitbucket.org/2.0/repositories/{self.workspace}/{self.repo_name}/src/{self.branch}/{key}"
